"""
Test functions to verify methods of BnrOneA class
"""
import os
import time
from one import BnrOneA


def scroll_text(text, size_of_line):
    """
    From the initial text and a specified size,
    yields a new text each time starting with an empty text
    and then shifting the text to the left
    ending with an empty text again
    :param text: input text
    :param size_of_line: size of the output text
    """
    extended_text = (" " * size_of_line) + text + (" " * size_of_line)
    for i in range(17 + len(text)):
        text = extended_text[i : i + 16]
        yield text


def test_scroll_text():
    """
    Sends scrolling text to the robot
    Text should be displayed in lcd line 2
    """
    print("=== Testing LCD scrolling text ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0

    text = "Hi Raspberry Pi!"
    for text in scroll_text(text, 16):
        print(text, end="\n")
        one.lcd2(text)
        time.sleep(0.2)
    print(" ")


def test_move():
    """
    Test move method
    """
    print("=== Testing move ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    delay_s = 1
    print("Move forward")
    one.move(30, 30)
    time.sleep(delay_s)
    print("Rotate right")
    one.move(30, -30)
    time.sleep(delay_s)
    print("Rotate left")
    one.move(-30, 30)
    time.sleep(delay_s)
    print("Move backwards")
    one.move(-30, -30)
    time.sleep(delay_s)
    print("Stop")
    one.move(0, 0)


def test_move_calibrate():
    """
    Test move calibrate method
    """
    print("=== Testing move calibrate ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    delay_ms = 2000
    for i in range(0, 101, 5):
        print("Duty_cycle:", i)
        one.move_calibrate(i, i)
        time.sleep(0.5)
    print("Stopping both motors")
    one.move_calibrate(0, 0)
    print("At the start motors are not supposed to move.")
    print("After a certain value they should start moving but not necessarily at the same time.")


def test_move_1m():
    """
    Test move one motor
    """
    print("=== Testing move one motor ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    left_motor = 1
    right_motor = 2
    speed = 30
    delay_s = 2
    print("Left wheel forward")
    one.move_1m(left_motor, speed)
    time.sleep(delay_s)
    print("Right wheel forward")
    one.move_1m(right_motor, speed)
    time.sleep(delay_s)
    print("Left wheel: STOP")
    one.move_1m(left_motor, 0)
    time.sleep(delay_s)
    print("Right wheel backwards")
    one.move_1m(right_motor, -speed)
    time.sleep(delay_s)
    print("Left wheel backwards")
    one.move_1m(left_motor, -speed)
    time.sleep(delay_s)
    print("Right wheel: STOP")
    one.move_1m(right_motor, 0)
    time.sleep(delay_s)
    print("Left wheel: STOP")
    one.move_1m(left_motor, 0)


def test_stop():
    """
    Test stop
    """
    print("=== Testing stop ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    speed = 30
    delay_s = 2
    print("Move forward")
    one.move(speed, speed)
    time.sleep(delay_s)
    print("STOP")
    one.stop()
    time.sleep(delay_s)
    print("Move backwards")
    one.move(-speed, -speed)
    time.sleep(delay_s)
    print("STOP")
    one.stop()
    time.sleep(delay_s)


def test_stop_1m():
    """
    Test stop 1 motor
    """
    print("=== Testing stop 1 motor ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    left_motor = 1
    right_motor = 2
    speed = 30
    delay_s = 2
    print("Move forward")
    one.move(speed, speed)
    time.sleep(delay_s)
    print("Left wheel: STOP")
    one.stop_1m(left_motor)
    time.sleep(delay_s)
    print("Right wheel: STOP")
    one.stop_1m(right_motor)
    time.sleep(delay_s)
    print("Move backwards")
    one.move(-speed, -speed)
    time.sleep(delay_s)
    print("Left wheel: STOP")
    one.stop_1m(left_motor)
    time.sleep(delay_s)
    print("Right wheel: STOP")
    one.stop_1m(right_motor)
    time.sleep(delay_s)


def test_brake():
    """
    Test brake
    """
    print("=== Testing brake ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    speed = 30
    delay_s = 2
    print("Move forward")
    one.move(speed, speed)
    time.sleep(delay_s)
    print("BRAKE 10%")
    one.brake(10, 10)
    time.sleep(delay_s)
    print("Move backwards")
    one.move(-speed, -speed)
    time.sleep(delay_s)
    print("BRAKE 70%")
    one.brake(70, 70)


def test_brake_1m():
    """
    Test brake 1 motor
    """
    print("=== Testing brake 1 motor ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    left_motor = 1
    right_motor = 2
    speed = 30
    delay_s = 2
    print("Move forward")
    one.move(speed, speed)
    time.sleep(delay_s)
    print("BRAKE 70% left motor")
    one.brake_1m(left_motor, 70)
    time.sleep(delay_s)
    print("BRAKE 70% right motor")
    one.brake_1m(right_motor, 70)
    time.sleep(delay_s)
    print("Move backwards")
    one.move(-speed, -speed)
    time.sleep(delay_s)
    print("BRAKE Default torque left motor")
    one.brake_1m(left_motor)
    time.sleep(delay_s)
    print("BRAKE Default torque right motor")
    one.brake_1m(right_motor)


def reset_encoder(side):
    """
    reset encoder
    """
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    motor = side
    left_motor = 1
    speed = 20
    delay_s = 1
    print("Move forward")
    if motor == left_motor:
        one.reset_left_encoder()
    else:
        one.reset_right_encoder()
    one.move_1m(motor, speed)
    time.sleep(delay_s)
    if motor == left_motor:
        before_reseting = int(one.read_left_encoder())
        time.sleep(delay_s)
        one.reset_left_encoder()
        after_reseting = int(one.read_left_encoder())
    else:
        before_reseting = int(one.read_right_encoder())
        time.sleep(delay_s)
        one.reset_right_encoder()
        after_reseting = int(one.read_right_encoder())
    one.stop_1m(motor)
    print("before_reseting:", before_reseting, "\tafter_reseting:", after_reseting)
    assert after_reseting < before_reseting
    assert after_reseting < 2


def test_reset_left_encoder():
    """
    Test reset left encoder
    """
    print("=== Testing reset left encoder ===")
    left_motor = 1
    reset_encoder(left_motor)


def test_reset_right_encoder():
    """
    Test reset right encoder
    """
    print("=== Testing reset right encoder ===")
    right_motor = 2
    reset_encoder(right_motor)


def blink(one, number_of_times, delay_s):
    """
    Blinks the led on Bot'n Roll One A

    :param one: object to send the command to
    :param number_of_times: number of times to toggle the led
    :duration between toggling
    """
    state = 1
    for i in range(1, number_of_times + 1):
        state = i % 2
        print("LED: ", state)
        one.led(state)
        time.sleep(delay_s)


def test_led():
    """
    Test led
    """
    print("=== Testing led blinking ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    blink(one, 6, 1)


def test_obstacle_emitters():
    """
    Test obstacle emitters
    """
    print("=== Testing obstacle emitters ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    one.obstacle_emitters(1)
    state = 0
    number_of_times = 6
    delay_s = 2
    print("Place obstacle close in front of the obstacle sensors")
    print("Verify red leds light on when emitter state is 1 and turn off otherwise")
    for i in range(0, number_of_times + 1):
        state = i % 2
        print("Emitter state: ", state)
        one.obstacle_emitters(state)
        time.sleep(delay_s)


def servo(id):
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    delay_s = 0.1
    print("Rotate in one way")
    for i in range(10, 170, 10):
        if id == 1:
            one.servo1(i)
        else:
            one.servo2(i)
        time.sleep(delay_s)
    print("Rotate in opposite way")
    for i in range(170, 10, -10):
        if id == 1:
            one.servo1(i)
        else:
            one.servo2(i)
        time.sleep(delay_s)
    one.servo1(90)


def test_servo1():
    """
    Test servo 1
    """
    print("=== Testing servo 1 ===")
    servo(1)


def test_servo2():
    """
    Test servo 2
    """
    print("=== Testing servo 2 ===")
    servo(2)


def test_min_battery():
    """
    Test min battery
    """
    print("=== Testing min battery ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Setting to a normal value.")
    print("Robot should move in response to a move command.")
    one.min_battery(8.4)
    print("voltage = ", one.read_battery())
    one.move(30, 30)
    time.sleep(2)
    print("Setting to a high value. LCD should display '* LOW BATTERY *'.")
    print("Robot should not move in response to a move command.")
    one.min_battery(15.0)
    print("voltage = ", one.read_battery())
    one.move(-20, -20)
    time.sleep(2)
    one.stop()
    print("!!! You need to reboot the robot to reestablish moving capabilities !!!")
    one.min_battery(8.4)


def test_save_calibrate_1():
    """
    Test save calibrate - 1st test
    """
    print("=== Testing save calibrate - 1st test===")
    print("!!! You will need to re-calibrate the motors after this test as calibration is lost !!!")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("1st time running this test the wheels are expected to move.")
    print("2nd time running this test the wheels should NOT move.")
    one.move(30, 30)
    time.sleep(2)
    one.stop()
    print("Setting minimum duty cycle as 0.")
    print("If this is the 1st time, please reboot the robot and run same test again.")
    one.save_calibrate(12, 5, 5)


def test_save_calibrate_2():
    """
    Test save calibrate - 2nd test
    """
    print("=== Testing save calibrate - 2nd test ===")
    print("!!! You will need to re-calibrate the motors after this test as calibration is lost !!!")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print(
        "1st time running this test: if test_save_calibrate_1 was run before this one the wheels should not move otherwise they will."
    )
    print("2nd time running this test the wheels should definitely move.")
    one.move(30, 30)
    time.sleep(2)
    one.stop()
    print("Setting minimum duty cycle as 0.")
    print("If this is the 1st time, please reboot the robot and run same test again.")
    one.save_calibrate(12, 70, 70)


def test_read_button():
    """
    Reads button pressed from the robot
    Note: User should press the buttons on the robot
    """
    print("=== Testing read button ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0

    print("Please press a button on the robot")
    time.sleep(2)
    for i in range(30):
        print("Test", i + 1, "of 30. Button pressed: ", one.read_button())
        time.sleep(0.3)


def test_read_battery():
    """
    Test voltage battery reading
    """
    print("=== Testing read battery ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Battery voltage = ", one.read_battery())


def read_encoder(side):
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    motor = side
    left_motor = 1
    speed = 20
    delay_s = 1
    print("Move forward")
    if motor == left_motor:
        one.reset_left_encoder()
    else:
        one.reset_right_encoder()
    one.move_1m(motor, speed)
    time.sleep(delay_s)
    if motor == left_motor:
        first_reading = int(one.read_left_encoder())
        time.sleep(delay_s * 1.1)
        second_reading = int(one.read_left_encoder())
        time.sleep(delay_s * 0.9)
        third_reading = int(one.read_left_encoder())
    else:
        first_reading = int(one.read_right_encoder())
        time.sleep(delay_s * 1.1)
        second_reading = int(one.read_right_encoder())
        time.sleep(delay_s * 0.9)
        third_reading = int(one.read_right_encoder())
    one.stop_1m(motor)
    print("first_reading:", first_reading, "\tsecond_reading:", second_reading, "\tthird_reading:", third_reading)
    assert third_reading < first_reading < second_reading


def test_read_left_encoder():
    """
    Test read left encoder
    """
    print("=== Testing read left encoder ===")
    left_motor = 1
    read_encoder(left_motor)


def test_read_right_encoder():
    """
    Test read right encoder
    """
    print("=== Testing read right encoder ===")
    right_motor = 2
    read_encoder(right_motor)


def read_encoder_increment(side):
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    motor = side
    left_motor = 1
    speed = 20
    delay_s = 1
    print("Move forward")
    if motor == left_motor:
        one.reset_left_encoder()
    else:
        one.reset_right_encoder()
    one.move_1m(motor, speed)
    time.sleep(delay_s)
    if motor == left_motor:
        first_reading = int(one.read_left_encoder_increment())
        time.sleep(delay_s * 1.1)
        second_reading = int(one.read_left_encoder_increment())
        time.sleep(delay_s * 0.9)
        third_reading = int(one.read_left_encoder_increment())
    else:
        first_reading = int(one.read_right_encoder_increment())
        time.sleep(delay_s * 1.1)
        second_reading = int(one.read_right_encoder_increment())
        time.sleep(delay_s * 0.9)
        third_reading = int(one.read_right_encoder_increment())
    one.stop_1m(motor)
    print("first_reading:", first_reading, "\tsecond_reading:", second_reading, "\tthird_reading:", third_reading)
    assert first_reading < second_reading < third_reading


def test_read_left_encoder_increment():
    """
    Test read left encoder increment
    """
    print("=== Testing read left encoder increment ===")
    left_motor = 1
    read_encoder_increment(left_motor)


def test_read_right_encoder_increment():
    """
    Test read right encoder increment
    """
    print("=== Testing read right encoder increment ===")
    right_motor = 2
    read_encoder_increment(right_motor)


def test_read_firmware():
    """
    Test read firmware
    """
    print("=== Testing read firmware ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    firmware = one.read_firmware()
    print("Firmware = ", firmware[0], ".", firmware[1], ".", firmware[2])


def test_obstacle_sensors():
    """
    Test obstacle sensors
    """
    print("=== Testing obstacle sensors ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    for i in range(100):
        print("Test", i + 1, "of 100: Output:", one.obstacle_sensors())
        time.sleep(0.2)


def test_read_IR_sensors():
    """
    Test read IR sensors
    """
    print("=== Testing read IR sensors ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    for i in range(100):
        print("Test", i + 1, "of 100: Output:", one.read_ir_sensors())
        time.sleep(0.2)


def test_read_left_range():
    """
    Test read left range sensor
    """
    print("=== Testing read left range sensor ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    for i in range(100):
        print("Test", i + 1, "of 100: Output:", one.read_left_range())
        time.sleep(0.2)


def test_read_right_range():
    """
    Test read right range sensor
    """
    print("=== Testing read right range sensor ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    for i in range(100):
        print("Test", i + 1, "of 100: Output:", one.read_right_range())
        time.sleep(0.2)


def test_read_adc():
    """
    Test read adc
    """
    print("=== Testing read adc ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    for i in range(8):
        print("Read adc", i, ":", one.read_adc(i))
        time.sleep(0.2)


def test_read_adc_0():
    """
    Test read adc 0
    """
    print("=== Testing read adc 0 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 0:", one.read_adc_0())


def test_read_adc_1():
    """
    Test read adc 1
    """
    print("=== Testing read adc 1 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 1:", one.read_adc_1())


def test_read_adc_2():
    """
    Test read adc 2
    """
    print("=== Testing read adc 2 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 2:", one.read_adc_2())


def test_read_adc_3():
    """
    Test read adc 3
    """
    print("=== Testing read adc 3 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 3:", one.read_adc_3())


def test_read_adc_4():
    """
    Test read adc 4
    """
    print("=== Testing read adc 4 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 4:", one.read_adc_4())


def test_read_adc_5():
    """
    Test read adc 5
    """
    print("=== Testing read adc 5 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 5:", one.read_adc_5())


def test_read_adc_6():
    """
    Test read adc 6
    """
    print("=== Testing read adc 6 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 6:", one.read_adc_6())


def test_read_adc_7():
    """
    Test read adc 7
    """
    print("=== Testing read adc 7 ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    print("Read adc 7:", one.read_adc_7())


def test_read_DBG():
    """
    Test read DBG
    """
    print("=== Testing read DBG ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0
    for i in range(4):
        print("Read dbg", i, ":", one.read_dbg(i))


def test_lcd():
    """
    Sends different types of text using both lines of the lcd
    User should verify the output by looking at the lcd on the robot
    """
    print("=== Testing writing data to LCD ===")
    one = BnrOneA(0, 0)  # creates a BotnRoll interface at bus 0 and channel 0

    delay_s = 1.6
    one.lcd1("")
    one.lcd2("Hi Raspberry Pi!")
    time.sleep(delay_s)
    one.lcd1("Hi Raspberry Pi!")
    one.lcd2("Day:", 31, "7", 2023)
    time.sleep(delay_s)
    one.lcd1("Day:", 31, "7", 2023)
    one.lcd2(17, "h", 15, "min")
    time.sleep(delay_s)
    one.lcd1(17, "h", 15, "min")
    one.lcd2("Ver.", 1, "Sub.", 3)
    time.sleep(delay_s)
    one.lcd1("Ver.", 1, "Sub.", 3)
    one.lcd2("Test number:", 1)
    time.sleep(delay_s)
    one.lcd1("Test number:", 1)
    one.lcd2("System", "test:", 1)
    time.sleep(delay_s)
    one.lcd1("System", "test:", 1)
    one.lcd2(1234567890123456)
    time.sleep(delay_s)
    one.lcd1(1234567890123456)
    one.lcd2(12345678, 1234567)
    time.sleep(delay_s)
    one.lcd1(12345678, 1234567)
    one.lcd2(12345, 12345, 1234)
    time.sleep(delay_s)
    one.lcd1(12345, 12345, 1234)
    one.lcd2(1111, 2222, 3333, 4444)
    time.sleep(delay_s)
    one.lcd1(1111, 2222, 3333, 4444)
    one.lcd2("      END       ")
    time.sleep(delay_s)
    one.lcd1("      END       ")
    one.lcd2("")


def main():
    """
    Calls functions to test public interface with BotnRoll One A
    Most of these tests should be verified with the robot connected to the raspbery pi
    and by visually inspecting the robot and/or the terminal
    """
    print("Run tests using: pytest", os.path.basename(__file__), "-s")

    # test_move()
    # test_move_calibrate()
    # test_move_1m()
    # test_stop()
    # test_stop_1m()
    # test_brake()
    # test_brake_1m()
    # test_reset_left_encoder()
    # test_reset_right_encoder()
    # test_led()
    # test_obstacle_emitters()
    # test_servo1()
    # test_servo2()
    # test_save_calibrate_1() # requires manual reboot
    # test_save_calibrate_2()  # requires manual reboot
    # test_min_battery() # requires manual reboot
    # test_read_button()
    # test_read_battery()
    # test_read_left_encoder()
    # test_read_right_encoder()
    # test_read_left_encoder_increment()
    # test_read_right_encoder_increment()
    # test_read_firmware()
    # test_obstacle_sensors()
    # test_read_IR_sensors()
    # test_read_left_range()
    # test_read_right_range()
    # test_read_adc()
    # test_read_adc_0()
    # test_read_adc_1()
    # test_read_adc_2()
    # test_read_adc_3()
    # test_read_adc_4()
    # test_read_adc_5()
    # test_read_adc_6()
    # test_read_adc_7()
    # test_read_DBG()
    # test_lcd()
    # test_scroll_text()


if __name__ == "__main__":
    main()



"""
Config module to store variables for the operation of BotnRoll One A.
Some variables are read from config files.
Other variables are computed from values read from config files.
These variables work like static variables.
You can read, modify them in the code and then save the relevant ones
in a file for next time.
"""

import json
import os


class Config:
    """
    Saves and load config values to and from file
    """

    sensor_min = [0] * 8  # array of 8 elements with min value for each line sensor
    sensor_max = [1000] * 8  # array of 8 elements with max value for each line sensor
    threshold = 50
    correction_factor = 6
    cfg_file = None

    def __init__(self, filename=None):
        if filename is None:
            file_name = "config.json"
            self.cfg_file = os.path.join(os.path.dirname(__file__), file_name)
        else:
            self.cfg_file = filename

    def load(self):
        with open(self.cfg_file) as f:
            data = json.load(f)

        # Access values from the JSON file
        self.sensor_min = data["sensor_min"]
        self.sensor_max = data["sensor_max"]
        self.threshold = data["threshold"]
        self.correction_factor = data["correction_factor"]

    def print(self):
        print("sensor_max:", self.sensor_max)
        print("sensor_min:", self.sensor_min)
        print("threshold:", self.threshold)
        print("correction_factor:", self.correction_factor)

    def save(self):
        # Save the updated dictionary back to the JSON file
        data = {
            "sensor_max": self.sensor_max,
            "sensor_min": self.sensor_min,
            "threshold": self.threshold,
            "correction_factor": self.correction_factor,
        }

        with open(self.cfg_file, "w", encoding="utf-8") as file:
            json.dump(data, file, indent=4)


def main():
    cfg = Config()
    cfg.save()
    cfg.load()
    cfg.print()
    cfg.save()
    cfg.print()


if __name__ == "__main__":
    main()




import sys
import os

# these steps are necessary in order to import modules
current_dir = os.path.dirname(os.path.abspath(__file__))
project_root = os.path.abspath(os.path.join(current_dir, ".."))
sys.path.append(project_root)

from utils.config import Config


def test_config():
    """
    Test config class and it's methods namely:
     - loading values from config file
     - saving values to config file
    """
    cfg = Config("test_cfg.json")
    print(">> Create config file")
    cfg.sensor_min = [100, 200, 300, 300]
    cfg.sensor_max = [500, 400, 700, 800]
    cfg.threshold = 18
    cfg.print()
    print(">> Saving config values into file...")
    cfg.save()
    cfg.print()
    print(">> Loading config values from file...")
    cfg.load()
    cfg.print()
    assert cfg.sensor_min == [100, 200, 300, 300]
    assert cfg.sensor_max == [500, 400, 700, 800]
    assert cfg.threshold == 18
    print(">> Changing values directly in config...")
    cfg.sensor_min = [100, 100, 100, 100]
    cfg.sensor_max = [200, 200, 200, 200]
    cfg.threshold = 22
    cfg.print()
    print(">> Assert that after loading values changed back")
    cfg.load()
    cfg.print()
    assert cfg.sensor_min == [100, 200, 300, 300]
    assert cfg.sensor_max == [500, 400, 700, 800]
    assert cfg.threshold == 18


def main():
    print("Run tests using: pytest", os.path.basename(__file__), "-s")

if __name__ == "__main__":
    main()


"""
Line detector class
"""

from utils.config import Config


class LineDetector:
    """
    Converts line sensor readings into a relative line position on the sensor.
    It assumes that the array of sensors of a sensor peripheral are arranged in a straight line and equally spaced.
    It provides methods to compute the location of the line relative to the device.
    It outputs values in a range [-100, 100] where 0 (zero) corresponds to the centre of the sensor peripheral.
    """

    _ref_max = 1000
    _cfg_loaded = (
        False  # flag to signal whether or not the config values have been read already
    )
    _scaling_factor = [
        0
    ] * 8  # array of 8 elements with correction factor for each line sensor
    _previous_line_value = 0
    _config = Config()

    def _normalise(self, reading, minimum, scale):
        """
        normalise a reading taking the minimum value of the range and a scale factor
        """
        return int((reading - minimum) * scale)

    def _normalise_readings(self, sensor_reading):
        """
        Normalize values for each sensor reading
        """
        # creating 'aliases' for config values to make code more readable
        sensor_min = self._config.sensor_min
        scaling_factor = (
            self._scaling_factor
        )  # this is a true alias as the object is mutable
        size = len(sensor_reading)
        sensor_normalised = [0] * size
        for i in range(size):
            sensor_normalised[i] = self._normalise(
                sensor_reading[i], sensor_min[i], scaling_factor[i]
            )
        return sensor_normalised

    def _calculate_factors(self, ref, min, max):
        """
        Calculates the scaling factor given a reference value\
        and a range defined by min and max values
        This scaling factor is useful in normalising values
        """
        factors = [ref] * len(min)
        divisor = [x - y for x, y in zip(max, min)]
        factors = [x / y for x, y in zip(factors, divisor)]
        return factors

    def _load_if_necessary(self):
        """
        Loads values from config if they weren't loaded before
        """
        if not self._cfg_loaded:
            self._config.load()
            self._scaling_factor = self._calculate_factors(
                self._ref_max, self._config.sensor_min, self._config.sensor_max
            )

    def _compute_line_value(self, readings):
        """
        Computes a line value in the range [0, ref_max]
        """
        line_value = -1
        max_reading = max(readings)
        if max_reading > self._config.threshold:
            line_value = self._compute_mean_gaussian(readings)
        return line_value

    def _cap_value(self, value, lower_limit, upper_limit):
        """
        Caps the value to lower and upper limits
        """
        if value < lower_limit:
            return lower_limit
        elif value > upper_limit:
            return upper_limit
        else:
            return value

    def _convert_range(self, x_value, x_min, x_max, y_min, y_max):
        """
        Converts a value x given in the range [x_min : x_max]
        to a new value in the range [y_min : y_max]
        """
        x_range = x_max - x_min
        y_range = y_max - y_min

        # Avoid division by zero
        if x_range == 0:
            return y_min + y_range / 2

        # Calculate the converted value
        y = ((x_value - x_min) / x_range) * y_range + y_min
        return y

    def _normalise_line_value(self, line_value, size):
        """
        Converts a line value in the range of [0, 8000] to a new range e.g. [-106, 106]
        This method involves two steps:
        1)  Converting the value in the range of [0, 8000] to a new range using a correction factor
            e.g. for a correction factor of 6% the new range is [-106, 106]
            The new range extends beyond 100 by a little margin specified by the correction factor in the config.
            The purpose of this is to enable adjusting the sensitivity of the readings near the extremities.
            For a sensor that is typically higher (away from the floor) this correction factor should be higher
            to prevent large jumps in the readings close to the extremities.
            By increasing the correction factor we are narrowing the sensitivity of the sensors and by reducing it
            we are extending the sensitivity out.
            There is a balance to achieve by tuning this value correctly in order to get the best of both worlds:
            - Smooth transitions in sensor value near the extremeties
            - Maximum sensitivity region of the sensor
        2) Capping the values to the desired range [-100, 100]
        """

        x_min = 0
        x_max = size * self._ref_max  # should be 8000
        y_min = -100 - self._config.correction_factor
        y_max = 100 + self._config.correction_factor
        line_value = self._convert_range(line_value, x_min, x_max, y_min, y_max)
        line_value = self._cap_value(line_value, -100, 100)
        return line_value

    def _filter_line_value(self, line_value, ref_value, max_value):
        """
        Filters the line value to handle edge cases
        such as no line detected or reading errors

        :param line_value: the computed line value
        :param ref_value: the reference value to help determining which side
        is the line in case it's no longer detected (e.g. middle of the range)
        :param max_value: the maximum value of the range
        """
        # out of the line -> all white
        if line_value == -1:
            if self._previous_line_value > ref_value:
                line_value = max_value
            else:
                line_value = 0
        # possible reading errors
        elif line_value < -1 or line_value > max_value:
            line_value = self._previous_line_value
        # if normal values
        else:
            self._previous_line_value = line_value
        return line_value

    def _compute_mean_gaussian(self, reading):
        """
        Lets assume the line detected gives us a discrete gaussian
        where the probabilities are given by each sensor reading and
        the values are pre-determined based on each sensor location:
         |sensor id | value  | probability |
         |----------|--------|-------------|
         |    0     |   1    |  reading[0] |
         |    1     |   2    |  reading[1] |
         |  (...)   | (...)  |    (...)    |
         |    7     |   8    |  reading[7] |

        We can compute the mean of the gaussian (location of line) by:
         1) computing the product for each sensor: product = value * probability
         2) computing sum_products = product[0] + product[1] + ... + product[7]
         3) computing sum_probabilities = reading[0] + reading[1] + ... + reading[7]
         4) computing mean = sum_products / sum_probabilities
        """
        size = len(reading)
        value = list(range(500, (size + 1) * 1000, 1000))
        product = [x * y for x, y in zip(value, reading)]
        sum_product = sum(product)
        sum_probability = sum(reading)
        if sum_probability == 0:
            mean = 0
        else:
            mean = sum_product / sum_probability
        return mean

    def _get_max_value_and_index(self, lst):
        """
        Returns the max value and its corresponding index from a list
        """
        max_value = max(lst)
        max_index = lst.index(max_value)
        return max_value, max_index

    def _prune(self, readings):
        """
        Deals with edge cases such when the max readings is on one of the sensors at the extremety
        In such cases it bumps up the value of the sensor to be at the threshold level
        """
        max_value, max_index = self._get_max_value_and_index(readings)
        if max_value < self._config.threshold:
            if max_index == 0 or max_index == (len(readings) - 1):
                readings[max_index] = self._config.threshold
        return readings

    def compute_line(self, readings):
        """
        Given an input as a set of sensors readings it computes a relative location of a line
        along the length of the sensor and expresses the output in a range [-100, 100]
        where 0 (zero) corresponds to the line being at the centre of the sensor.
        """
        max_range = len(readings) * 1000
        mid_range = max_range / 2
        self._load_if_necessary()
        normalised = self._normalise_readings(readings)
        pruned = self._prune(normalised)
        line_value = self._compute_line_value(pruned)
        line_value = self._filter_line_value(line_value, mid_range, max_range)
        line_value = self._normalise_line_value(line_value, len(pruned))
        line_value = self._cap_value(line_value, -100, 100)
        return line_value



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

name = "onepi"
version = "0.1.0"
description = "Python library to interface with BotnRoll One A"
url = "https://github.com/ninopereira/bnronepi/tree/main/onepi"
author = "Nino Pereira"
author_email = "ninopereira.pt@gmail.com"
license = "MIT"
packages = ["onepi"]
classifiers = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]

# Package dependencies
install_requires = ["spidev"]

setup(
    name=name,
    version=version,
    description=description,
    url=url,
    author=author,
    author_email=author_email,
    license=license,
    classifiers=classifiers,
    packages=find_packages(),
    install_requires=install_requires,
)



"""
Test line detector
test_compute_line tests the public interface of the line detector class
When developing your own application you should only use the public methods.
All the other tests in this file are testing the private methods.
"""

import sys
import os
import time
import matplotlib.pyplot as plt
import numpy as np

# these steps are necessary in order to import modules
current_dir = os.path.dirname(os.path.abspath(__file__))
project_root = os.path.abspath(os.path.join(current_dir, ".."))
sys.path.append(project_root)

from utils.line_detector import LineDetector

plt.ion()


def test_normalise():
    line_detector = LineDetector()
    normalised_reading = line_detector._normalise(50, 20, 1.5)
    expected = 45
    assert normalised_reading == expected
    normalised_reading = line_detector._normalise(0, 20, 1.5)
    expected = -30
    assert normalised_reading == expected
    normalised_reading = line_detector._normalise(-20, 20, 1.5)
    expected = -60
    assert normalised_reading == expected
    normalised_reading = line_detector._normalise(20, 20, 1.5)
    expected = 0
    assert normalised_reading == expected

def test_normalise_readings():
    print("test_normalise_readings")
    line_detector = LineDetector()
    file_name = "test_cfg.json"
    cfg_file = os.path.join(os.path.dirname(__file__), file_name)
    line_detector._config.cfg_file = cfg_file
    line_detector._load_if_necessary()  # depends on config values
    readings = [222, 333, 444, 555]
    normalised_readings = line_detector._normalise_readings(readings)
    expected = [305, 665, 360, 510]
    assert normalised_readings == expected


def test_load_if_necessary():
    print("test_load_if_necessary")
    line_detector = LineDetector()
    file_name = "test_cfg.json"
    cfg_file = os.path.join(os.path.dirname(__file__), file_name)
    line_detector._config.cfg_file = cfg_file
    line_detector._load_if_necessary()  # depends on config values
    # after loading it should have the
    # min_values, max_values and scaling factors
    assert line_detector._config.sensor_min == [100, 200, 300, 300]
    assert line_detector._config.sensor_max == [500, 400, 700, 800]
    assert line_detector._scaling_factor == [2.5, 5.0, 2.5, 2.0]


def test_compute_line_value():
    print("test_compute_line_value")
    line_detector = LineDetector()
    # first we need to load config values
    file_name = "test_cfg.json"
    cfg_file = os.path.join(os.path.dirname(__file__), file_name)
    line_detector._config.cfg_file = cfg_file
    line_detector._load_if_necessary()
    readings = [0, 0, 700, 0]
    line = line_detector._compute_line_value(readings)
    assert line == 2500


def test_cap_value():
    print("test_cap_value")
    line_detector = LineDetector()
    # test below min
    capped = line_detector._cap_value(10, 20, 100)
    assert capped == 20
    # test above max
    capped = line_detector._cap_value(130, 20, 100)
    assert capped == 100
    # test within range
    capped = line_detector._cap_value(60, 20, 100)
    assert capped == 60


def test_convert_range():
    print("test_convert_range")
    line_detector = LineDetector()
    converted = line_detector._convert_range(20, 0, 100, -100, 0)
    assert converted == -80


def test_normalise_line_value():
    print("test_normalise_line_value")
    line_detector = LineDetector()
    input_values = [0, 4000, 8000]
    expected_values = [-100, 0, 100]
    for value, expected in zip(input_values, expected_values):
        normalised_value = int(line_detector._normalise_line_value(value, 8))
        assert normalised_value == expected


def test_filter_line_value():
    print("test_filter_line_value")
    line_detector = LineDetector()
    # first we need to load config values
    file_name = "test_cfg.json"
    cfg_file = os.path.join(os.path.dirname(__file__), file_name)
    line_detector._config.cfg_file = cfg_file
    line_detector._load_if_necessary()
    # test valid reading
    filtered = line_detector._filter_line_value(200, 2000, 4000)
    assert filtered == 200
    # test no line detected
    filtered = line_detector._filter_line_value(-1, 2000, 4000)
    assert filtered == 0
    # test invalid value
    line_detector._filter_line_value(3600, 2000, 4000)
    filtered = line_detector._filter_line_value(-10, 2000, 4000)
    assert filtered == 3600
    # test invalid value above max
    line_detector._filter_line_value(2400, 2000, 4000)
    filtered = line_detector._filter_line_value(6000, 2000, 4000)
    assert filtered == 2400
    # test no line detected
    filtered = line_detector._filter_line_value(3700, 2000, 4000)
    filtered = line_detector._filter_line_value(-1, 2000, 4000)
    assert filtered == 4000


def test_compute_mean_gaussian():
    print("test_compute_mean_gaussian")
    line_detector = LineDetector()
    sensor_readings = [10] * 8
    sensor_readings[2] = 600
    line = line_detector._compute_mean_gaussian(sensor_readings)
    assert int(line) == 2679


def test_get_max_value_and_index():
    print("test_get_max_value_and_index")
    line_detector = LineDetector()
    readings = [4, 8, 6, 2]
    max_value, index = line_detector._get_max_value_and_index(readings)
    assert max_value == 8
    assert index == 1


def test_prune():
    print("test_prune")
    line_detector = LineDetector()
    # first we need to load config values
    file_name = "test_cfg.json"
    cfg_file = os.path.join(os.path.dirname(__file__), file_name)
    line_detector._config.cfg_file = cfg_file
    line_detector._load_if_necessary()
    # test normal case where max is not on the extremety
    readings = [4, 8, 6, 2]
    pruned = line_detector._prune(readings)
    assert pruned == [4, 8, 6, 2]
    # test edge case where max is at the start
    readings = [8, 3, 6, 2]
    pruned = line_detector._prune(readings)
    assert pruned == [18, 3, 6, 2]
    # test edge case where max is at the end
    readings = [8, 3, 6, 12]
    pruned = line_detector._prune(readings)
    assert pruned == [8, 3, 6, 18]


def test_compute_line():
    """
    Test the public interface compute_line
    """
    print("test_compute_line")
    line_detector = LineDetector()
    # first we need to set config values
    line_detector._config.sensor_min = [0] * 4
    line_detector._config.sensor_max = [1000] * 4
    line_detector._scaling_factor = [1] * 4
    line_detector._cfg_loaded = True
    # No line present
    sensor_readings = [10] * 4
    line = line_detector.compute_line(sensor_readings)
    assert line < -99
    # Line at the centre
    sensor_readings[1:3] = [800, 800]
    line = line_detector.compute_line(sensor_readings)
    assert -10 < line < 10
    # Line on the right side
    sensor_readings = [10] * 4
    sensor_readings[2] = 800
    line = line_detector.compute_line(sensor_readings)
    assert 20 < line < 30
    # Line not found
    sensor_readings = [10] * 4
    line = line_detector.compute_line(sensor_readings)
    assert line == 100
    # Line on the left side
    sensor_readings = [10] * 4
    sensor_readings[1] = 800
    line = line_detector.compute_line(sensor_readings)
    assert line < 50


def main():
    print("Run tests using: pytest", os.path.basename(__file__), "-s")


if __name__ == "__main__":
    main()
